#!/usr/bin/env python3
"""Command-line tool for plotting magnon DOS from TB2J results."""

import argparse
from pathlib import Path

import numpy as np

from TB2J.magnon.magnon3 import Magnon
from TB2J.magnon.magnon_dos import plot_magnon_dos


def main():
    parser = argparse.ArgumentParser(
        description="Calculate and plot magnon DOS from TB2J results"
    )
    parser.add_argument(
        "-p",
        "--path",
        default="TB2J_results",
        help="Path to TB2J results directory (default: TB2J_results)",
    )
    parser.add_argument(
        "--kmesh",
        type=int,
        nargs=3,
        default=[20, 20, 20],
        metavar=("nx", "ny", "nz"),
        help="k-point mesh dimensions (default: 20, 20, 20)",
    )
    parser.add_argument(
        "--no-gamma",
        action="store_false",
        dest="gamma",
        help="Exclude Gamma point from k-mesh",
    )
    parser.add_argument(
        "--width",
        type=float,
        default=0.001,
        help="Gaussian smearing width in eV (default: 0.001)",
    )
    parser.add_argument(
        "--window",
        type=float,
        nargs=2,
        metavar=("emin", "emax"),
        help="Energy window in meV (optional)",
    )
    parser.add_argument(
        "--npts",
        type=int,
        default=401,
        help="Number of energy points (default: 401)",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="magnon_dos.png",
        help="Output filename for plot (default: magnon_dos.png)",
    )
    parser.add_argument(
        "-s",
        "--show",
        action="store_true",
        dest="show",
        help="Show plot window",
    )

    # Exchange interaction options (same as in magnon bands)
    parser.add_argument(
        "-j",
        "--Jiso",
        action="store_true",
        default=True,
        help="Include isotropic exchange interactions (default: True)",
    )
    parser.add_argument(
        "--no-Jiso",
        action="store_false",
        dest="Jiso",
        help="Exclude isotropic exchange interactions",
    )
    parser.add_argument(
        "-a",
        "--Jani",
        action="store_true",
        default=False,
        help="Include anisotropic exchange interactions (default: False)",
    )
    parser.add_argument(
        "-d",
        "--DMI",
        action="store_true",
        default=False,
        help="Include Dzyaloshinskii-Moriya interactions (default: False)",
    )

    # Reference vector options (same as in magnon bands)
    parser.add_argument(
        "-q",
        "--Q",
        nargs=3,
        type=float,
        metavar=("Qx", "Qy", "Qz"),
        help="Propagation vector [Qx, Qy, Qz] (default: [0, 0, 0])",
    )
    parser.add_argument(
        "-u",
        "--uz-file",
        type=str,
        help="Path to file containing quantization axes for each spin (nspin×3 array)",
    )
    parser.add_argument(
        "-c",
        "--spin-conf-file",
        type=str,
        help="Path to file containing magnetic moments for each spin (nspin×3 array)",
    )
    parser.add_argument(
        "-v",
        "--n",
        nargs=3,
        type=float,
        metavar=("nx", "ny", "nz"),
        help="Normal vector for rotation [nx, ny, nz] (default: [0, 0, 1])",
    )

    args = parser.parse_args()

    # Check if TB2J results exist
    if not Path(args.path).exists():
        raise FileNotFoundError(f"TB2J results not found at {args.path}")

    # Load magnon calculator with exchange interaction options
    print(f"Loading exchange parameters from {args.path}...")
    magnon = Magnon.from_TB2J_results(
        path=args.path, Jiso=args.Jiso, Jani=args.Jani, DMI=args.DMI
    )

    # Set reference vectors if provided (same logic as in magnon bands)
    Q = [0, 0, 0] if args.Q is None else args.Q
    n = [0, 0, 1] if args.n is None else args.n

    # Handle quantization axes
    if args.uz_file is not None:
        # Make path relative to TB2J results if not absolute
        uz_file = args.uz_file
        if not Path(uz_file).is_absolute():
            uz_file = str(Path(args.path) / uz_file)

        uz = np.loadtxt(uz_file)
        if uz.shape[1] != 3:
            raise ValueError(
                f"Quantization axes file should contain a nspin×3 array. Got shape {uz.shape}"
            )
        if uz.shape[0] != magnon.nspin:
            raise ValueError(
                f"Number of spins in uz file ({uz.shape[0]}) does not match the system ({magnon.nspin})"
            )
    else:
        # Default: [0, 0, 1] for all spins
        uz = np.array([[0, 0, 1]], dtype=float)

    # Handle spin configuration
    if args.spin_conf_file is not None:
        # Make path relative to TB2J results if not absolute
        spin_conf_file = args.spin_conf_file
        if not Path(spin_conf_file).is_absolute():
            spin_conf_file = str(Path(args.path) / spin_conf_file)

        magmoms = np.loadtxt(spin_conf_file)
        if magmoms.shape[1] != 3:
            raise ValueError(
                f"Spin configuration file should contain a nspin×3 array. Got shape {magmoms.shape}"
            )
        if magmoms.shape[0] != magnon.nspin:
            raise ValueError(
                f"Number of spins in spin configuration file ({magmoms.shape[0]}) does not match the system ({magnon.nspin})"
            )
    else:
        magmoms = None

    # Set reference configuration
    magnon.set_reference(Q, uz, n, magmoms)

    # Convert window from meV to eV if provided
    window = None
    if args.window is not None:
        window = (args.window[0] / 1000, args.window[1] / 1000)

    # Calculate and plot DOS
    print("\nCalculating magnon DOS...")
    _dos = plot_magnon_dos(
        magnon,
        kmesh=args.kmesh,
        gamma=args.gamma,
        width=args.width,
        window=window,
        npts=args.npts,
        filename=args.output,
        show=args.show,
    )

    print(f"\nPlot saved to {args.output}")
    data_file = Path(args.output).with_suffix(".json")
    print(f"DOS data saved to {data_file}")


if __name__ == "__main__":
    main()



"""
Module about Pauli (and I) matrices.
"""

import numpy as np
from numpy import zeros_like
from scipy.linalg import svd

s0 = np.array([[1, 0], [0, 1]])
s1 = np.array([[0, 1], [1, 0]])
s2 = np.array([[0, -1j], [1j, 0]])
s3 = np.array([[1, 0], [0, -1]])

s0T = s0.T
s1T = s1.T
s2T = s2.T
s3T = s3.T

pauli_dict = {0: s0, 1: s1, 2: s2, 3: s3}


def pauli_mat(nbasis, i):
    """
    nbasis: size of the matrix. should be multiple of 2.
    i: index of pauli dictionary.
    """
    N = nbasis // 2
    assert N * 2 == nbasis
    M = np.ones((N, N), dtype="complex")
    spm = pauli_dict[i]
    return np.block([[M * spm[0, 0], M * spm[0, 1]], [M * spm[1, 0], M * spm[1, 1]]])


def pauli_decomp(M):
    """Given a 2*2 matrix, get the I, x, y, z component.
    :param M: 2*2 matrix
    :returns:  (I, x, y, z) are four scalars.
    :rtype: same as dtype of M
    """
    return (
        np.trace(s0.dot(M)) / 2,
        np.trace(s1.dot(M)) / 2,
        np.trace(s2.dot(M)) / 2,
        np.trace(s3.dot(M)) / 2,
    )


def pauli_decomp2(M):
    """Given a 2*2 matrix, get the I, x, y, z component. (method2)
    :param M: 2*2 matrix
    :returns:  (I, x, y, z) are four scalars.
    :rtype: same as dtype of M
    """
    return (
        np.sum(M * s0T) / 2,
        np.sum(M * s1T) / 2,
        np.sum(M * s2T) / 2,
        np.sum(M * s3T) / 2,
    )


def pauli_sigma_norm_old(M):
    MI, Mx, My, Mz = pauli_decomp2(M)
    return np.linalg.norm([Mx, My, Mz])


def pauli_block_I(M, norb):
    """
    I compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, ::2] + M[1::2, 1::2]) / 2
    ret[::2, ::2] = ret[1::2, 1::2] = tmp
    return ret


def pauli_block_x(M, norb):
    """
    x compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, 1::2] + M[1::2, ::2]) / 2
    ret[::2, 1::2] = ret[1::2, ::2] = tmp
    return ret


def pauli_block_y(M, norb):
    """
    y compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, 1::2] * (-1j) + M[1::2, ::2] * (1j)) / 2
    ret[::2, 1::2] = tmp * (-1j)
    ret[1::2, ::2] = tmp * 1j
    return tmp, ret


def pauli_block_z(M, norb):
    """z compoenent of a matrix, see pauli_block
    :param M:
    :param norb:
    :returns:
    :rtype:
    """
    ret = zeros_like(M)
    tmp = (M[::2, ::2] - M[1::2, 1::2]) / 2
    ret[::2, ::2] = tmp
    ret[1::2, 1::2] = -tmp
    return tmp, ret


def pauli_block(M, idim):
    """Get the I, x, y, z component of a matrix.
    :param M: The input matrix,  aranged in four blocks:
    [[upup, updn], [dnup, dndn]]. let norb be number of orbitals in
    each block. (so M has dim 2norb1*2norb2)
    :param idim: 0, 1,2, 3 for I , x, y, z.
    :returns:  the idim(th) component of M
    :rtype: a matrix with shape of M.shape//2
    """
    # ret = zeros_like(M)
    if idim == 0:
        tmp = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    elif idim == 1:
        tmp = (M[::2, 1::2] + M[1::2, ::2]) / 2.0
    elif idim == 2:
        # Note that this is not element wise product with sigma_y but dot product
        # sigma_y=[[0, -1j],[1j, 0]]
        tmp = (M[::2, 1::2] * (1.0j) + M[1::2, ::2] * (-1.0j)) / 2.0
    elif idim == 3:
        tmp = (M[::2, ::2] - M[1::2, 1::2]) / 2.0
    else:
        raise NotImplementedError()
    return tmp


def pauli_block_all_old(M):
    MI = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    Mx = (M[::2, 1::2] + M[1::2, ::2]) / 2.0
    # Note that this is not element wise product with sigma_y but dot product
    My = (M[::2, 1::2] - M[1::2, ::2]) * 0.5j
    Mz = (M[::2, ::2] - M[1::2, 1::2]) / 2.0
    return MI, Mx, My, Mz


def pauli_block_all(array):
    A00 = array[..., ::2, ::2]
    A01 = array[..., ::2, 1::2]
    A10 = array[..., 1::2, ::2]
    A11 = array[..., 1::2, 1::2]
    n2 = array.shape[-1] // 2

    out_dtype = np.result_type(array.dtype, np.complex64)
    block = np.empty((*array.shape[:-2], 4, n2, n2), dtype=out_dtype)

    np.add(A00, A11, out=block[..., 0, :, :])
    block[..., 0, :, :] *= 0.5
    np.add(A01, A10, out=block[..., 1, :, :])
    block[..., 1, :, :] *= 0.5
    np.subtract(A01, A10, out=block[..., 2, :, :])
    block[..., 2, :, :] *= 0.5j
    np.subtract(A00, A11, out=block[..., 3, :, :])
    block[..., 3, :, :] *= 0.5

    return block


def gather_pauli_blocks(MI, Mx, My, Mz, coeffs=[1.0, 1.0, 1.0, 1.0]):
    """
    Gather the I, x, y, z component of a matrix.
    """
    cI, cx, cy, cz = coeffs
    return (
        cI * np.kron(MI, s0)
        + cx * np.kron(Mx, s1)
        + cy * np.kron(My, s2)
        + cz * np.kron(Mz, s3)
    )


def pauli_part(M, coeffs=[1.0, 1.0, 1.0, 1.0]):
    """
    Get the I, x, y, z part of a matrix.
    """
    MI, Mx, My, Mz = pauli_block_all(M)
    return gather_pauli_blocks(MI, Mx, My, Mz, coeffs=coeffs)


def chargepart(M):
    """
    Get the charge part of a matrix.
    """
    MI = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    Mcopy = np.zeros_like(M)
    Mcopy[::2, ::2] = MI
    Mcopy[1::2, 1::2] = MI
    return Mcopy


def spinpart(M):
    """
    Get the spin part of a matrix.
    """
    MI = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    Mcopy = M.copy()
    Mcopy[::2, ::2] -= MI
    Mcopy[1::2, 1::2] -= MI
    return Mcopy


def test_gather_pauli_blocks():
    M = np.random.rand(4, 4)
    MI, Mx, My, Mz = pauli_block_all(M)
    M2 = gather_pauli_blocks(MI, Mx, My, Mz)
    assert np.allclose(M, M2)


def op_norm(M):
    """
    Return the operator norm of a matrix.
    """
    return max(svd(M)[1])


def pauli_block_sigma_norm(M):
    """
    M= MI * I + \vec{P} dot \vec{sigma}
    = MI*I + p * \vec{e} dot \vec{sigma}
    where p is the norm of P.
    """
    MI, Mx, My, Mz = pauli_block_all(M)
    ex, ey, ez = np.trace(Mx), np.trace(My), np.trace(Mz)
    # ex,ey,ez = op_norm(Mx), op_norm(My), op_norm(Mz)
    evec = np.array((ex, ey, ez))
    evec = evec / np.linalg.norm(evec)
    return Mx * evec[0] + My * evec[1] + Mz * evec[2]


def pauli_sigma_norm(array):
    block = pauli_block_all(array)[..., 1:, :, :]
    E = np.trace(block, axis1=-2, axis2=-1)
    E /= np.linalg.norm(E, axis=-1, keepdims=True)
    np.multiply(block, E[..., None, None], out=block)
    return block.sum(axis=-3)



import numpy as np
from TB2J.abacus.abacus_wrapper import AbacusWrapper, AbacusParser
from TB2J.mathutils.rotate_spin import rotate_Matrix_from_z_to_axis
from TB2J.kpoints import monkhorst_pack
from TB2J.mathutils.fermi import fermi
from TB2J.mathutils.kR_convert import R_to_k
from scipy.linalg import eigh
from copy import deepcopy
from scipy.spatial.transform import Rotation
import matplotlib.pyplot as plt
from pathlib import Path
from TB2J.abacus.occupations import Occupations

# TODO List:
# - [x] Add the class AbacusSplitSOCWrapper
# - [x] Add the function to rotate the XC part
# - [x] Compute the band energy at arbitrary angle


def get_occupation(evals, kweights, nel, width=0.1):
    occ = Occupations(nel=nel, width=width, wk=kweights, nspin=2)
    return occ.occupy(evals)


def get_density_matrix(evals=None, evecs=None, kweights=None, nel=None, width=0.1):
    occ = get_occupation(evals, kweights, nel, width=width)
    rho = np.einsum("kib, kb, kjb -> kij", evecs, occ, evecs.conj())
    return rho


def spherical_to_cartesian(theta, phi, normalize=True):
    """
    Convert spherical coordinates to cartesian
    """
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    vec = np.array([x, y, z])
    if normalize:
        vec = vec / np.linalg.norm(vec)
    return vec


class AbacusSplitSOCWrapper(AbacusWrapper):
    """
    Abacus wrapper with Hamiltonian split to SOC and non-SOC parts
    """

    def __init__(self, *args, **kwargs):
        HR_soc = kwargs.pop("HR_soc", None)
        # nbasis = HR_soc.shape[1]
        # kwargs["nbasis"] = nbasis
        super().__init__(*args, **kwargs)
        self._HR_copy = deepcopy(self._HR)
        self.HR_soc = HR_soc
        self.soc_lambda = 1.0
        self.nel = 16
        self.width = 0.1

    @property
    def HR(self):
        return self._HR + self.HR_soc * self.soc_lambda

    def rotate_HR_xc(self, axis):
        """
        Rotate SOC part of Hamiltonian
        """
        for iR, R in enumerate(self.Rlist):
            self._HR[iR] = rotate_Matrix_from_z_to_axis(self._HR_copy[iR], axis)

    def rotate_Hk_xc(self, axis):
        """
        Rotate SOC part of Hamiltonian
        """
        for ik in range(len(self._Hk)):
            self._Hk[ik] = rotate_Matrix_from_z_to_axis(self._Hk_copy[ik], axis)

    def get_density_matrix(self, kpts, kweights=None):
        rho = np.zeros((len(kpts), self.nbasis, self.nbasis), dtype=complex)
        evals, evecs = self.solve_all(kpts)
        occ = get_occupation(evals, kweights, self.nel, width=self.width)
        rho = np.einsum(
            "kib, kb, kjb -> kij", evecs, occ, evecs.conj()
        )  # should multiply S to the the real DM.
        return rho

    def rotate_DM(self, rho, axis):
        """
        Rotate the density matrix
        """
        for ik in range(len(rho)):
            rho[ik] = rotate_Matrix_from_z_to_axis(rho[ik], axis)
        return rho


class RotateHam:
    def __init__(self, model, kmesh, gamma=True):
        self.model = model
        self.kpts = monkhorst_pack(kmesh, gamma_center=gamma)
        self.kweights = np.ones(len(self.kpts), dtype=float) / len(self.kpts)

    def get_band_energy2(self):
        for ik, kpt in enumerate(self.kpts):
            Hk, Sk = self.model.gen_ham(kpt)
            evals, evecs = eigh(Hk, Sk)
            rho = np.einsum(
                "ib, b, jb -> ij",
                evecs,
                fermi(evals, self.model.efermi, width=0.05),
                evecs.conj(),
            )
            eband1 = np.sum(evals * fermi(evals, self.model.efermi, width=0.05))
            eband2 = np.trace(Hk @ rho)
            print(eband1, eband2)

    def get_band_energy(self, dm=False):
        evals, evecs = self.model.solve_all(self.kpts)
        occ = get_occupation(
            evals, self.kweights, self.model.nel, width=self.model.width
        )
        eband = np.sum(evals * occ * self.kweights[:, np.newaxis])
        # * fermi(evals, self.model.efermi, width=0.05)
        if dm:
            density_matrix = self.model.get_density_matrix(evecs)
            return eband, density_matrix
        else:
            return eband

    def calc_ref(self):
        # calculate the Hk_ref, Sk_ref, Hk_soc_ref, and rho_ref
        self.Sk_ref = R_to_k(self.kpts, self.model.Rlist, self.model.SR)
        self.Hk_xc_ref = R_to_k(self.kpts, self.model.Rlist, self.model._HR_copy)
        self.Hk_soc_ref = R_to_k(self.kpts, self.model.Rlist, self.model.HR_soc)
        self.rho_ref = np.zeros(
            (len(self.kpts), self.model.nbasis, self.model.nbasis), dtype=complex
        )

        evals = np.zeros((len(self.kpts), self.model.nbasis), dtype=float)
        evecs = np.zeros(
            (len(self.kpts), self.model.nbasis, self.model.nbasis), dtype=complex
        )

        for ik, kpt in enumerate(self.kpts):
            # evals, evecs = eigh(self.Hk_xc_ref[ik]+self.Hk_soc_ref[ik], self.Sk_ref[ik])
            evals[ik], evecs[ik] = eigh(self.Hk_xc_ref[ik], self.Sk_ref[ik])
        occ = get_occupation(
            evals, self.kweights, self.model.nel, width=self.model.width
        )
        # occ = fermi(evals, self.model.efermi, width=self.model.width)
        self.rho_ref = np.einsum("kib, kb, kjb -> kij", evecs, occ, evecs.conj())

    def get_band_energy_from_rho(self, axis):
        """
        This is wrong!! Should use second order perturbation theory to get the band energy instead.
        """
        eband = 0.0
        for ik, k in enumerate(self.kpts):
            rho = rotate_Matrix_from_z_to_axis(self.rho_ref[ik], axis)
            Hk_xc = rotate_Matrix_from_z_to_axis(self.Hk_xc_ref[ik], axis)
            Hk_soc = self.Hk_soc_ref[ik]
            Htot = Hk_xc + Hk_soc * self.model.soc_lambda
            # Sk = self.Sk_ref[ik]
            # evals, evecs = eigh(Htot, Sk)
            # rho2= np.einsum("ib, b, jb -> ij", evecs, fermi(evals, self.model.efermi, width=0.05), evecs.conj())
            if ik == 0 and False:
                pass
                # print(f"{evecs[:4,0:4].real=}")
                # print(f"{evals[:4]=}")
                # print(f"{Hk_xc[:4,0:4].real=}")
                # print(f"{Htot[:4,0:4].real=}")
                # print(f"{Sk[:4,0:4].real=}")
                # print(f"{rho[:4,0:4].real=}")
                # print(f"{rho2[:4,0:4].real=}")
            # eband1 = np.sum(evals * fermi(evals, self.model.efermi, width=0.05))
            # eband2 = np.trace(Htot @ rho2).real
            # eband3 = np.trace(Htot @ rho).real
            # print(eband1, eband2, eband3)
            e_soc = np.trace(Hk_soc @ rho) * self.kweights[ik] * self.model.soc_lambda
            eband += e_soc
        return eband

    def get_band_energy_vs_angles(
        self,
        thetas,
        psis,
    ):
        es = []
        # es2 = []
        # e,rho = self.model.get_band_energy(dm=True)
        # self.calc_ref()
        # thetas = np.linspace(*angle_range, npoints)
        for i, theta, phi in enumerate(zip(thetas, psis)):
            axis = spherical_to_cartesian(theta, phi)
            self.model.rotate_HR_xc(axis)
            # self.get_band_energy2()
            e = self.get_band_energy()
            es.append(e)
            # es2.append(e2)
        return es


def get_model_energy(model, kmesh, gamma=True):
    ham = RotateHam(model, kmesh, gamma=gamma)
    return ham.get_band_energy()


class AbacusSplitSOCParser:
    """
    Abacus parser with Hamiltonian split to SOC and non-SOC parts
    """

    def __init__(self, outpath_nosoc=None, outpath_soc=None, binary=False):
        self.outpath_nosoc = outpath_nosoc
        self.outpath_soc = outpath_soc
        self.binary = binary
        self.parser_nosoc = AbacusParser(outpath=outpath_nosoc, binary=binary)
        self.parser_soc = AbacusParser(outpath=outpath_soc, binary=binary)
        spin1 = self.parser_nosoc.read_spin()
        spin2 = self.parser_soc.read_spin()
        if spin1 != "noncollinear" or spin2 != "noncollinear":
            raise ValueError("Spin should be noncollinear")

    def parse(self):
        nbasis, Rlist, HR, SR = self.parser_nosoc.Read_HSR_noncollinear()
        nbasis2, Rlist2, HR2, SR2 = self.parser_soc.Read_HSR_noncollinear()
        # print(HR[0])
        HR_soc = HR2 - HR
        model = AbacusSplitSOCWrapper(HR, SR, Rlist, nbasis, nspin=2, HR_soc=HR_soc)
        model.efermi = self.parser_soc.efermi
        model.basis = self.parser_nosoc.basis
        model.atoms = self.parser_nosoc.atoms
        return model


def abacus_get_MAE(
    path_nosoc, path_soc, kmesh, thetas, psis, gamma=True, outfile="MAE.txt"
):
    """Get MAE from Abacus with magnetic force theorem. Two calculations are needed. First we do an calculation with SOC but the soc_lambda is set to 0. Save the density. The next calculatin we start with the density from the first calculation and set the SOC prefactor to 1. With the information from the two calcualtions, we can get the band energy with magnetic moments in the direction, specified in two list, thetas, and phis."""
    parser = AbacusSplitSOCParser(
        outpath_nosoc=path_nosoc, outpath_soc=path_soc, binary=False
    )
    model = parser.parse()
    ham = RotateHam(model, kmesh, gamma=gamma)
    es = []
    for theta, psi in zip(thetas, psis):
        axis = spherical_to_cartesian(theta, psi)
        model.rotate_HR_xc(axis)
        e = ham.get_band_energy()
        es.append(ham.get_band_energy())
    if outfile:
        with open(outfile, "w") as f:
            f.write("theta, psi, energy\n")
            for theta, psi, e in zip(thetas, psis, es):
                f.write(f"{theta}, {psi}, {e}\n")
    return es


def test_AbacusSplitSOCWrapper():
    # path = Path("~/projects/2D_Fe").expanduser()
    path = Path("~/projects/TB2Jflows/examples/2D_Fe/Fe_z").expanduser()
    outpath_nosoc = f"{path}/soc0/OUT.ABACUS"
    outpath_soc = f"{path}/soc1/OUT.ABACUS"
    parser = AbacusSplitSOCParser(
        outpath_nosoc=outpath_nosoc, outpath_soc=outpath_soc, binary=False
    )
    model = parser.parse()
    kmesh = [6, 6, 1]

    r = RotateHam(model, kmesh)
    # thetas, es = r.get_band_energy_vs_theta(angle_range=(0, np.pi*2), rotation_axis="z", initial_direction=(1,0,0),  npoints=21)
    thetas, es, es2 = r.get_band_energy_vs_theta(
        angle_range=(0, np.pi * 2),
        rotation_axis="y",
        initial_direction=(0, 0, 1),
        npoints=11,
    )
    # print the table of thetas and es, es2
    print("theta, e, e2")
    for theta, e, e2 in zip(thetas, es, es2):
        print(f"{theta=}, {e=}, {e2=}")

    plt.plot(thetas / np.pi, es - es[0], marker="o")
    plt.plot(thetas / np.pi, es2 - es2[0], marker=".")
    plt.savefig("E_along_z_x_z.png")
    plt.show()


def abacus_get_MAE_cli():
    import argparse

    parser = argparse.ArgumentParser(
        description="Get MAE from Abacus with magnetic force theorem. Two calculations are needed. First we do an calculation with SOC but the soc_lambda is set to 0. Save the density. The next calculatin we start with the density from the first calculation and set the SOC prefactor to 1. With the information from the two calcualtions, we can get the band energy with magnetic moments in the direction, specified in two list, thetas, and phis. "
    )
    parser.add_argument("path_nosoc", type=str, help="Path to the  calculation with ")
    parser.add_argument("path_soc", type=str, help="Path to the SOC calculation")
    parser.add_argument("thetas", type=float, nargs="+", help="Thetas")
    parser.add_argument("psis", type=float, nargs="+", help="Phis")
    parser.add_argument("kmesh", type=int, nargs=3, help="K-mesh")
    parser.add_argument(
        "--gamma", action="store_true", help="Use Gamma centered kpoints"
    )
    parser.add_argument(
        "--outfile",
        type=str,
        help="The angles and the energey will be saved in this file.",
    )
    args = parser.parse_args()
    abacus_get_MAE(
        args.path_nosoc,
        args.path_soc,
        args.kmesh,
        args.thetas,
        args.psis,
        gamma=args.gamma,
        outfile=args.outfile,
    )


if __name__ == "__main__":
    abacus_get_MAE_cli()



"""Test basic loading and distance calculation for EPW data.

PURPOSE:
    Verify the distance_resolved_couplings methods work without hanging.

USAGE:
    uv run python agent_files/debug_epw_distance/test_basic.py
"""

import os
import sys
import time

# Add parent directory to path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

from HamiltonIO.epw.epwparser import Epmat, read_crystal_fmt
from ase.units import Bohr


def main():
    # Find test data
    base = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    data_dir = os.path.join(base, "HamiltonIO", "epw", "test", "up")

    print(f"Loading data from: {data_dir}")

    # Read crystal structure
    t0 = time.time()
    crystal = read_crystal_fmt(os.path.join(data_dir, "crystal.fmt"))
    print(f"Crystal loaded in {time.time() - t0:.2f}s")
    print(f"  natom={crystal.natom}, nmode={crystal.nmode}")

    # Read EPW data
    t0 = time.time()
    ep = Epmat()
    ep.crystal = crystal
    ep.read(path=data_dir, prefix="SrMnO3", epmat_ncfile="epmat.nc")
    print(f"Epmat loaded in {time.time() - t0:.2f}s")
    if hasattr(ep, "epmatfile") and ep.epmatfile is not None:
        print(f"  nwann={ep.nwann}, using netCDF file")
    elif hasattr(ep, "epmat_wann") and ep.epmat_wann is not None:
        print(f"  nwann={ep.nwann}, shape={ep.epmat_wann.shape}")
    else:
        print(f"  nwann={ep.nwann}, no matrix data loaded!")

    # Get cell
    cell = crystal.at.reshape(3, 3) * crystal.alat * Bohr
    print(f"Cell (Angstrom):\n{cell}")

    # Test Rk-based distance calculation
    imode = 0
    print(f"\nTesting Rk-based distances for mode {imode}...")
    t0 = time.time()
    entries_Rk = ep.distance_resolved_couplings_Rk(imode=imode, cell=cell)
    print(f"Rk calculation done in {time.time() - t0:.2f}s")
    print(f"  Found {len(entries_Rk)} entries")
    if entries_Rk:
        print(f"  Sample entry: {entries_Rk[0]}")
        # Print distance and magnitude statistics
        distances = [e["distance"] for e in entries_Rk]
        mags = [abs(e["coupling"]) for e in entries_Rk]
        print(f"  Distance range: {min(distances):.2f} - {max(distances):.2f} Angstrom")
        print(f"  |g| range: {min(mags):.4e} - {max(mags):.4e} eV/Angstrom")

    # Test Rg-based distance calculation
    print(f"\nTesting Rg-based distances for mode {imode}...")
    t0 = time.time()
    entries_Rg = ep.distance_resolved_couplings_Rg(imode=imode, cell=cell)
    print(f"Rg calculation done in {time.time() - t0:.2f}s")
    print(f"  Found {len(entries_Rg)} entries")
    if entries_Rg:
        print(f"  Sample entry: {entries_Rg[0]}")
        distances = [e["distance"] for e in entries_Rg]
        mags = [abs(e["coupling"]) for e in entries_Rg]
        print(f"  Distance range: {min(distances):.2f} - {max(distances):.2f} Angstrom")
        print(f"  |g| range: {min(mags):.4e} - {max(mags):.4e} eV/Angstrom")

    print("\nSuccess!")


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import numpy as np


class BandsPlot:
    _UNITS = "meV"
    _NSYSTEMS = 1

    def __init__(self, bands, kpath, xlist=None, **kwargs):
        """Initialize BandsPlot.

        Parameters
        ----------
        bands : array_like
            Band energies
        kpath : list of (index, label) tuples
            K-point labels and their indices
        xlist : list of arrays, optional
            X-coordinates for segmented paths. If None, uses range(nbands)
        **kwargs : dict
            Additional plotting options
        """
        self.bands = bands * 1000  # Convert to meV
        self.kpath = kpath
        self.xlist = xlist

        plot_options = kwargs
        self.linewidth = plot_options.pop("linewidth", 1.5)
        self.color = plot_options.pop("color", "blue")
        self.fontsize = plot_options.pop("fontsize", 12)
        self.ticksize = plot_options.pop("ticksize", 10)
        self.plot_options = plot_options

    def plot(self, filename=None):
        """Plot the band structure.

        Parameters
        ----------
        filename : str, optional
            If provided, saves the plot to this file
        """
        fig, axs = plt.subplots(1, self._NSYSTEMS, constrained_layout=True)

        # Get min/max for y-axis limits
        bmin, bmax = self.bands.min(), self.bands.max()
        ymin = bmin - 0.05 * np.abs(bmin - bmax)
        ymax = bmax + 0.05 * np.abs(bmax - bmin)
        axs.set_ylim([ymin, ymax])

        # Plot bands
        if self.xlist is not None:
            # Plot segments
            start_idx = 0
            for x in self.xlist:
                nbands = x.shape[0]
                segment_bands = self.bands[start_idx : start_idx + nbands].T
                for band in segment_bands:
                    axs.plot(
                        x,
                        band[start_idx : start_idx + nbands],
                        linewidth=self.linewidth,
                        color=self.color,
                        **self.plot_options,
                    )
                start_idx += nbands
            # Set xlim to cover all segments
            axs.set_xlim([self.xlist[0][0], self.xlist[-1][-1]])
        else:
            # Standard continuous plotting
            kdata = np.arange(self.bands.shape[0])
            for band in self.bands.T:
                axs.plot(
                    kdata,
                    band,
                    linewidth=self.linewidth,
                    color=self.color,
                    **self.plot_options,
                )
            axs.set_xlim([0, kdata[-1]])

        # Add k-point labels and vertical lines
        kpoint_labels = list(zip(*self.kpath))
        axs.set_xticks(kpoint_labels[0], kpoint_labels[1], fontsize=self.ticksize)
        axs.vlines(
            x=kpoint_labels[0],
            ymin=ymin,
            ymax=ymax,
            color="black",
            linewidth=self.linewidth / 5,
        )

        axs.set_ylabel(f"Energy ({self._UNITS})", fontsize=self.fontsize)

        if filename is None:
            plt.show()
        else:
            plt.savefig(filename, dpi=300, bbox_inches="tight")



from dataclasses import asdict, dataclass
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import tomli
import tomli_w
from ase.dft.dos import DOS
from ase.units import J, eV
from scipy.spatial.transform import Rotation

from TB2J.io_exchange import SpinIO
from TB2J.kpoints import monkhorst_pack
from TB2J.magnon.magnon_band import MagnonBand
from TB2J.magnon.magnon_math import get_rotation_arrays
from TB2J.mathutils.auto_kpath import auto_kpath


@dataclass
class MagnonParameters:
    """Parameters for magnon band structure calculations"""

    path: str = "TB2J_results"
    kpath: str = None
    npoints: int = 300
    filename: str = "magnon_bands.png"
    Jiso: bool = True
    Jani: bool = False
    DMI: bool = False
    Q: Optional[List[float]] = None
    uz_file: Optional[str] = None
    n: Optional[List[float]] = None
    spin_conf_file: Optional[str] = None
    show: bool = False

    @classmethod
    def from_toml(cls, filename: str) -> "MagnonParameters":
        """Load parameters from a TOML file"""
        with open(filename, "rb") as f:
            data = tomli.load(f)
        return cls(**data)

    def to_toml(self, filename: str):
        """Save parameters to a TOML file"""
        # Convert to dict and remove None values
        data = {k: v for k, v in asdict(self).items() if v is not None}
        with open(filename, "wb") as f:
            tomli_w.dump(data, f)

    def __post_init__(self):
        """Validate parameters after initialization"""
        if self.Q is not None and len(self.Q) != 3:
            raise ValueError("Q must be a list of 3 numbers")
        if self.n is not None and len(self.n) != 3:
            raise ValueError("n must be a list of 3 numbers")

        # Convert path to absolute path if uz_file is relative to it
        if self.uz_file and not Path(self.uz_file).is_absolute():
            self.uz_file = str(Path(self.path) / self.uz_file)
        if self.spin_conf_file and not Path(self.spin_conf_file).is_absolute():
            self.spin_conf_file = str(Path(self.path) / self.spin_conf_file)


@dataclass
class Magnon:
    """
    Magnon calculator implementation using dataclass
    """

    nspin: int
    # ind_atoms: list
    magmom: np.ndarray
    Rlist: np.ndarray
    JR: np.ndarray
    cell: np.ndarray
    _Q: np.ndarray
    _uz: np.ndarray
    _n: np.ndarray
    pbc: tuple = (True, True, True)

    def set_reference(self, Q, uz, n, magmoms=None):
        """
        Set reference propagation vector and quantization axis

        Parameters
        ----------
        Q : array_like
            Propagation vector
        uz : array_like
            Quantization axis
        n : array_like
            Normal vector for rotation
        """
        self.set_propagation_vector(Q)
        self._uz = np.array(uz, dtype=float)
        self._n = np.array(n, dtype=float)
        if magmoms is not None:
            self.magmom = np.array(magmoms, dtype=float)

    def set_propagation_vector(self, Q):
        """Set propagation vector"""
        self._Q = np.array(Q)

    @property
    def Q(self):
        """Get propagation vector"""
        if self._Q is None:
            raise ValueError("Propagation vector Q is not set.")
        return self._Q

    @Q.setter
    def Q(self, value):
        if not isinstance(value, (list, np.ndarray)):
            raise TypeError("Propagation vector Q must be a list or numpy array.")
        if len(value) != 3:
            raise ValueError("Propagation vector Q must have three components.")
        self._Q = np.array(value)

    def Jq(self, kpoints):
        """
        Compute the exchange interactions in reciprocal space.

        The exchange interactions J(q) are computed using the Fourier transform:
        J(q) = ∑_R J(R) exp(iq·R)

        Array shapes and indices:
        - kpoints: (nkpt, 3) array of k-points
        - Rlist: (nR, 3) array of real-space lattice vectors
        - JR: (nR, nspin, nspin, 3, 3) array of exchange tensors in real space
            where nspin is number of magnetic atoms
        - Output Jq: (nkpt, nspin, nspin, 3, 3) array of exchange tensors in q-space

        If propagation vector Q is set, each J(R) is rotated before the Fourier transform:
        J'_mn(R) = R_m(ϕ)^T J(R) R_n(ϕ)
        where ϕ = 2π R·Q and R(ϕ) is rotation matrix around axis n by angle ϕ

        Parameters
        ----------
        kpoints : array_like (nkpt, 3)
            k-points at which to evaluate the exchange interactions

        Returns
        -------
        numpy.ndarray (nkpt, nspin, nspin, 3, 3)
            Exchange interaction tensors J(q) at each k-point
            First two indices are for magnetic atom pairs
            Last two indices are for 3x3 tensor components
        """
        Rlist = np.array(self.Rlist)
        JR = self.JR
        JRprime = JR.copy()

        for iR, R in enumerate(Rlist):
            if self._Q is not None:
                # Rotate exchange tensors based on propagation vector
                phi = 2 * np.pi * R @ self._Q  # angle ϕ = 2π R·Q
                rv = phi * self._n  # rotation vector
                Rmat = Rotation.from_rotvec(rv).as_matrix()
                # J'_mn(R) = R_m(ϕ)^T J(R) R_n(ϕ) using Einstein summation.
                # Here m is always in the R=0, thus the rotation is only applied on the
                # n , so only on the right.
                JRprime[iR] = np.einsum(" ijxy, yb -> ijxb", JR[iR], Rmat)

        nkpt = kpoints.shape[0]
        Jq = np.zeros((nkpt, self.nspin, self.nspin, 3, 3), dtype=complex)

        for iR, R in enumerate(Rlist):
            for iqpt, qpt in enumerate(kpoints):
                # Fourier transform of exchange tensors
                phase = 2 * np.pi * R @ qpt
                Jq[iqpt] += np.exp(1j * phase) * JRprime[iR]

        # Jq_copy = Jq.copy()
        # Jq.swapaxes(-1, -2)  # swap xyz
        # Jq.swapaxes(-3, -4)  # swap ij
        # Jq = (Jq.conj() + Jq_copy) / 2.0
        return Jq

    def Hq(self, kpoints):
        """
        Compute the magnon Hamiltonian in reciprocal space.

        Parameters
        ----------
        kpoints : array_like
            k-points at which to evaluate the Hamiltonian
        anisotropic : bool, optional
            Whether to include anisotropic interactions, default True

        Returns
        -------
        numpy.ndarray
            Magnon Hamiltonian matrix at each k-point
        """
        magmoms = self.magmom.copy()
        magmoms /= np.linalg.norm(magmoms, axis=-1)[:, None]

        U, V = get_rotation_arrays(magmoms, u=self._uz)

        J0 = -self.Jq(np.zeros((1, 3)))[0]
        # J0 = -Hermitize(J0)[:, :, 0]
        # Jq = -Hermitize(self.Jq(kpoints, anisotropic=anisotropic))

        Jq = -self.Jq(kpoints)

        C = np.diag(np.einsum("ix,ijxy,jy->i", V, 2 * J0, V))
        B = np.einsum("ix,kijxy,jy->kij", U, Jq, U)
        A1 = np.einsum("ix,kijxy,jy->kij", U, Jq, U.conj())
        A2 = np.einsum("ix,kijxy,jy->kij", U.conj(), Jq, U)

        H = np.block([[A1 - C, B], [B.swapaxes(-1, -2).conj(), A2 - C]])
        return H

    def _magnon_energies(self, kpoints, u=None):
        """Calculate magnon energies"""
        H = self.Hq(kpoints)
        n = H.shape[-1] // 2
        I = np.eye(n)

        min_eig = 0.0
        try:
            K = np.linalg.cholesky(H)
        except np.linalg.LinAlgError:
            try:
                K = np.linalg.cholesky(H + 1e-6 * np.eye(2 * n))
            except np.linalg.LinAlgError:
                from warnings import warn

                min_eig = np.min(np.linalg.eigvalsh(H))
                K = np.linalg.cholesky(H - (min_eig - 1e-6) * np.eye(2 * n))
                warn(
                    f"WARNING: The system may be far from the magnetic ground-state. Minimum eigenvalue: {min_eig}. The magnon energies might be unphysical."
                )

        g = np.block([[1 * I, 0 * I], [0 * I, -1 * I]])
        KH = K.swapaxes(-1, -2).conj()
        # Why only n:?
        return np.linalg.eigvalsh(KH @ g @ K)[:, n:] + min_eig
        # return np.linalg.eigvalsh(KH @ g @ K)[:, :] + min_eig

    def get_magnon_bands(
        self,
        kpoints: np.array = np.array([]),
        path: str = None,
        npoints: int = 300,
        special_points: dict = None,
        tol: float = 2e-4,
        pbc: tuple = None,
        cartesian: bool = False,
        labels: list = None,
        anisotropic: bool = True,
        u: np.array = None,
    ):
        """Get magnon band structure.

        Parameters
        ----------
        kpoints : np.array, optional
            Explicit k-points to calculate bands at. If empty, generates k-points from path.
        path : str, optional
            String specifying the k-path. If None, generates automatically using auto_kpath.
        npoints : int, optional
            Number of k-points along the path. Default is 300.
        special_points : dict, optional
            Dictionary of special points coordinates.
        tol : float, optional
            Tolerance for k-point comparisons. Default is 2e-4.
        pbc : tuple, optional
            Periodic boundary conditions. Default is None.
        cartesian : bool, optional
            Whether k-points are in cartesian coordinates. Default is False.
        labels : list, optional
            List of k-point labels. Default is None.
        anisotropic : bool, optional
            Whether to include anisotropic interactions. Default is True.
        u : np.array, optional
            Quantization axis. Default is None.

        Returns
        -------
        tuple
            - labels : list of (index, name) tuples for special k-points
            - bands : array of band energies
            - xlist : list of arrays with x-coordinates for plotting (if using auto_kpath)
        """
        pbc = self.pbc if pbc is None else pbc
        pbc = [True, True, True]
        u = self._uz if u is None else u
        if kpoints.size == 0:
            if path is None:
                # Use auto_kpath to generate path automatically
                xlist, kptlist, Xs, knames, spk = auto_kpath(
                    self.cell, None, npoints=npoints
                )
                kpoints = np.concatenate(kptlist)
                # Create labels from special points
                labels = []
                current_pos = 0
                for i, (x, k) in enumerate(zip(xlist, kptlist)):
                    for name in knames:
                        matches = np.where((k == spk[name]).all(axis=1))[0]
                        if matches.size > 0:
                            labels.append((matches[0] + current_pos, name))
                    current_pos += len(k)
            else:
                bandpath = self.cell.bandpath(
                    path=path,
                    npoints=npoints,
                    special_points=special_points,
                    eps=tol,
                    pbc=pbc,
                )
                kpoints = bandpath.kpts
                spk = bandpath.special_points
                spk[r"$\Gamma$"] = spk.pop("G", np.zeros(3))
                labels = [
                    (i, symbol)
                    for symbol in spk
                    for i in np.where((kpoints == spk[symbol]).all(axis=1))[0]
                ]
        elif cartesian:
            kpoints = np.linalg.solve(self.cell.T, kpoints.T).T

        bands = self._magnon_energies(kpoints)
        print(f"bands shape: {bands.shape}")

        if path is None and kpoints.size == 0:  # Fixed condition
            # When using auto_kpath, return xlist for segmented plotting
            return labels, bands, xlist
        else:
            return labels, bands, None

    def plot_magnon_bands(self, **kwargs):
        """
        Plot magnon band structure.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments passed to get_magnon_bands and plotting functions.
            Supported plotting options:
            - filename : str, optional
                Output filename for saving the plot
            - ax : matplotlib.axes.Axes, optional
                Axes for plotting. If None, creates new figure
            - show : bool, optional
                Whether to show the plot on screen
        """
        filename = kwargs.pop("filename", None)
        kpath_labels, bands, xlist = self.get_magnon_bands(**kwargs)

        # Get k-points and special points
        if "path" in kwargs and kwargs["path"] is None:
            _, kptlist, _, _, spk = auto_kpath(
                self.cell, None, npoints=kwargs.get("npoints", 300)
            )
            kpoints = np.concatenate(kptlist)
        else:
            bandpath = self.cell.bandpath(
                path=kwargs.get("path", "GXMG"), npoints=kwargs.get("npoints", 300)
            )
            kpoints = bandpath.kpts
            spk = bandpath.special_points.copy()
            spk[r"$\Gamma$"] = spk.pop("G", np.zeros(3))

        bands_plot = MagnonBand(
            energies=bands * 1000,  # Convert to meV
            kpoints=kpoints,
            kpath_labels=kpath_labels,
            special_points=spk,
            xcoords=xlist,
        )

        return bands_plot.plot(filename=filename, **kwargs)

    @classmethod
    def load_from_io(cls, exc: SpinIO, **kwargs):
        """
        Create Magnon instance from SpinIO

        Parameters
        ----------
        exc : SpinIO
            SpinIO instance with exchange parameters
        **kwargs : dict
            Additional arguments passed to get_full_Jtensor_for_Rlist

        Returns
        -------
        Magnon
            Initialized Magnon instance
        """
        # Get magnetic moments for magnetic atoms
        magmoms = exc.get_magnetic_moments()
        # nspin = len(magmoms)  # Number of magnetic atoms

        cell = exc.atoms.get_cell()
        pbc = exc.atoms.get_pbc()

        return cls(
            nspin=exc.nspin,
            magmom=magmoms,
            Rlist=exc.Rlist,
            JR=exc.get_full_Jtensor_for_Rlist(order="ij33", asr=False, **kwargs),
            cell=cell,
            _Q=np.zeros(3),  # Default propagation vector
            _uz=np.array([[0.0, 0.0, 1.0]]),  # Default quantization axis
            _n=np.array([0.0, 0.0, 1.0]),  # Default rotation axis
            pbc=pbc,
        )

    @classmethod
    def from_TB2J_results(cls, path=None, fname="TB2J.pickle", **kwargs):
        """
        Create Magnon instance from TB2J results.

        Parameters
        ----------
        path : str, optional
            Path to the TB2J results file
        fname : str, optional
            Filename of the TB2J results file, default "TB2J.pickle"
        **kwargs : dict
            Additional arguments passed to load_from_io

        Returns
        -------
        Magnon
            Initialized Magnon instance
        """
        exc = SpinIO.load_pickle(path=path, fname=fname)
        return cls.load_from_io(exc, **kwargs)


def test_magnon(path="TB2J_results"):
    """Test the magnon calculator by loading from TB2J_results and computing at high-symmetry points."""
    from pathlib import Path

    import numpy as np

    # Check if TB2J_results exists
    results_path = Path(path)
    if not results_path.exists():
        raise FileNotFoundError(f"TB2J_results directory not found at {path}")

    # Load magnon calculator from TB2J results
    print(f"Loading exchange parameters from {path}...")
    magnon = Magnon.from_TB2J_results(path=path, Jiso=True, Jani=False, DMI=False)

    # Define high-symmetry points for a cube
    kpoints = np.array(
        [
            [0.0, 0.0, 0.0],  # Γ (Gamma)
            [0.5, 0.0, 0.0],  # X
            [0.5, 0.5, 0.0],  # M
            [0.5, 0.5, 0.5],  # R
        ]
    )
    klabels = ["Gamma", "X", "M", "R"]

    print("\nComputing exchange interactions at high-symmetry points...")
    Jq = magnon.Jq(kpoints)

    print(f"\nResults for {len(kpoints)} k-points:")
    print("-" * 50)
    print("Exchange interactions J(q):")
    print(f"Shape of Jq tensor: {Jq.shape}")
    print(
        f"Dimensions: (n_kpoints={Jq.shape[0]}, n_spin={Jq.shape[1]}, n_spin={Jq.shape[2]}, xyz={Jq.shape[3]}, xyz={Jq.shape[4]})"
    )

    print("\nComputing magnon energies...")
    energies = magnon._magnon_energies(kpoints)

    print("\nMagnon energies at high-symmetry points (in meV):")
    print("-" * 50)
    for i, (k, label) in enumerate(zip(kpoints, klabels)):
        print(f"\n{label}-point k={k}:")
        # print(f"Energies: {energies[i] * 1000:.3f} meV")  # Convert to meV
        print(f"Energies: {energies[i] * 1000} meV")  # Convert to meV

    print("\nPlotting magnon bands...")
    magnon.plot_magnon_bands(
        # kpoints=kpoints,
        # labels=klabels,
        path="GHPGPH,PN",
        filename="magnon_bands.png",
    )

    return magnon, Jq, energies


def create_plot_script(filename: str):
    """Create a Python script for plotting the saved band structure data.

    Parameters
    ----------
    filename : str
        Base filename (without extension) to use for the plot script
    """
    script_name = f"plot_{filename}.py"
    script = '''#!/usr/bin/env python3
"""Simple script to plot magnon band structure from saved data."""

from TB2J.magnon.magnon_band import MagnonBand
import matplotlib.pyplot as plt

def plot_magnon_bands(input_file, output_file=None, ax=None, color='blue', show=True):
    """Load and plot magnon band structure.
    
    Parameters
    ----------
    input_file : str
        JSON file containing band structure data
    output_file : str, optional
        Output file for saving the plot
    ax : matplotlib.axes.Axes, optional
        Axes for plotting. If None, creates new figure
    color : str, optional
        Color of the band lines (default: blue)
    show : bool, optional
        Whether to show the plot on screen (default: True)
    
    Returns
    -------
    matplotlib.axes.Axes
        The plotting axes
    """
    # Load band structure data
    bands = MagnonBand.load(input_file)
    
    # Create plot
    ax = bands.plot(
        ax=ax,
        filename=output_file,
        color=color,
        show=show
    )
    return ax

if __name__ == "__main__":
    # Usage example
    # Example usage
    import matplotlib.pyplot as plt
    
    # Create a figure and axis (optional)
    fig, ax = plt.subplots(figsize=(6, 4))
    
    # Plot bands with custom color on given axis
    plot_magnon_bands(
        input_file="magnon_bands.json",
        output_file="magnon_bands.png",
        ax=ax,
        color='red',
        show=True
    )
'''

    with open(script_name, "w") as f:
        f.write(script)

    import os

    os.chmod(script_name, 0o755)  # Make executable


def save_bands_data(
    kpoints: np.ndarray,
    energies: np.ndarray,
    kpath_labels: List[Tuple[int, str]],
    special_points: dict,
    xcoords: Optional[Union[np.ndarray, List[np.ndarray]]],
    filename: str,
):
    """Save magnon band structure data to a JSON file using MagnonBand class.

    Parameters
    ----------
    kpoints : array_like
        Array of k-points coordinates
    energies : array_like
        Array of band energies (in meV)
    kpath_labels : list of (int, str)
        List of tuples containing k-point indices and their labels
    special_points : dict
        Dictionary of special points and their coordinates
    xcoords : array_like or list of arrays
        x-coordinates for plotting (can be segmented)
    filename : str
        Output filename
    """
    from TB2J.magnon.magnon_band import MagnonBand  # Using same import as above

    bands = MagnonBand(
        energies=energies,
        kpoints=kpoints,
        kpath_labels=kpath_labels,
        special_points=special_points,
        xcoords=xcoords,
    )
    bands.save(filename)

    # Create plotting script
    base_name = filename.rsplit(".", 1)[0]
    create_plot_script(base_name)

    print(f"Band structure data saved to {filename}")
    print(f"Created plotting script: plot_{base_name}.py")
    print("Usage: ")
    print(f"See plot_{base_name}.py for example usage")

    return bands


def plot_magnon_bands_from_TB2J(
    params: MagnonParameters,
):
    """
    Load TB2J results and plot magnon band structure along a specified k-path.

    Parameters
    ----------
    path : str, optional
        Path to TB2J results directory, default is "TB2J_results"
    kpath : str, optional
        String specifying the k-path, e.g. "GXMR" for Gamma-X-M-R path
        Default is "GXMR"
    npoints : int, optional
        Number of k-points along the path, default is 300
    filename : str, optional
        Output file name for the plot, default is "magnon_bands.png"
    Jiso : bool, optional
        Include isotropic exchange interactions, default is True
    Jani : bool, optional
        Include anisotropic exchange interactions, default is False
    DMI : bool, optional
        Include Dzyaloshinskii-Moriya interactions, default is False
    Q : array-like, optional
        Propagation vector [Qx, Qy, Qz], default is [0, 0, 0]
    uz_file : str, optional
        Path to file containing quantization axes for each spin (natom×3 array)
        If not provided, default [0, 0, 1] will be used for all spins
    n : array-like, optional
        Normal vector for rotation [nx, ny, nz], default is [0, 0, 1]
    show: bool, optional
        whether to show figure.

    Returns
    -------
    magnon : Magnon
        The Magnon instance used for calculations
    """
    # Load magnon calculator from TB2J results
    print(f"Loading exchange parameters from {params.path}...")
    magnon = Magnon.from_TB2J_results(
        path=params.path, Jiso=params.Jiso, Jani=params.Jani, DMI=params.DMI
    )

    # Set reference vectors if provided
    Q = [0, 0, 0] if params.Q is None else params.Q
    n = [0, 0, 1] if params.n is None else params.n

    # Handle quantization axes
    if params.uz_file is not None:
        uz = np.loadtxt(params.uz_file)
        if uz.shape[1] != 3:
            raise ValueError(
                f"Quantization axes file should contain a natom×3 array. Got shape {uz.shape}"
            )
        if uz.shape[0] != magnon.nspin:
            raise ValueError(
                f"Number of spins in uz file ({uz.shape[0]}) does not match the system ({magnon.nspin})"
            )
    else:
        # Default: [0, 0, 1] for all spins
        # uz = np.array([[0.0, 0.0, 1.0] for _ in range(magnon.nspin)])
        uz = np.array([[0, 0, 1]], dtype=float)

    print(params)
    if params.spin_conf_file is not None:
        magmoms = np.loadtxt(params.spin_conf_file)
        if magmoms.shape[1] != 3:
            raise ValueError(
                f"Spin configuration file should contain a nspin×3 array. Got shape {magmoms.shape}"
            )
        if magmoms.shape[0] != magnon.nspin:
            raise ValueError(
                f"Number of spins in spin configuration file ({magmoms.shape[0]}) does not match the system ({magnon.nspin})"
            )
    else:
        magmoms = None

    magnon.set_reference(Q, uz, n, magmoms)

    # Get band structure data
    print(f"\nCalculating bands along path {params.kpath}...")
    kpath_labels, bands, xlist = magnon.get_magnon_bands(
        path=params.kpath,
        npoints=params.npoints,
    )

    # Convert energies to meV
    bands_meV = bands * 1000

    # Save band structure data and create plot
    data_file = params.filename.rsplit(".", 1)[0] + ".json"
    print(f"\nSaving band structure data to {data_file}")

    # Get k-points and special points
    if params.kpath is None:
        _, kptlist, _, _, spk = auto_kpath(magnon.cell, None, npoints=params.npoints)
        kpoints = np.concatenate(kptlist)
    else:
        bandpath = magnon.cell.bandpath(path=params.kpath, npoints=params.npoints)
        kpoints = bandpath.kpts
        spk = bandpath.special_points
        spk[r"$\Gamma$"] = spk.pop("G", np.zeros(3))

    magnon_bands = save_bands_data(
        kpoints=kpoints,
        energies=bands_meV,
        kpath_labels=kpath_labels,
        special_points=spk,
        xcoords=xlist,
        filename=data_file,
    )

    # Plot band structure
    print(f"Plotting bands to {params.filename}")
    magnon_bands.plot(filename=params.filename)

    return magnon


def plot_magnon_bands_cli():
    import argparse
    import warnings

    warnings.warn(
        """ 
        # !!!!!!!!!!!!!!!!!! WARNING: =============================
        # 
        # This functionality is under development and should not be used in production.
        # It is provided for testing and development purposes only.
        # Please use with caution and report any issues to the developers.
        #
        # This warning will be removed in future releases.
        # =====================================

        """,
        UserWarning,
        stacklevel=2,
    )
    parser = argparse.ArgumentParser(
        description="Plot magnon band structure from TB2J results"
    )

    # Add a mutually exclusive group for config file vs. command line arguments
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--config",
        type=str,
        help="Path to TOML configuration file",
    )
    group.add_argument(
        "--save-config",
        type=str,
        help="Save default configuration to specified TOML file",
    )

    # Command line arguments (used if no config file is provided)
    parser.add_argument(
        "-p",
        "--path",
        default="TB2J_results",
        help="Path to TB2J results directory (default: TB2J_results)",
    )
    parser.add_argument(
        "-k",
        "--kpath",
        default=None,
        help="k-path specification (default: auto-detected from type of cell)",
    )
    parser.add_argument(
        "-n",
        "--npoints",
        type=int,
        default=300,
        help="Number of k-points along the path (default: 300)",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="magnon_bands.png",
        help="Output file name (default: magnon_bands.png)",
    )
    parser.add_argument(
        "-j",
        "--Jiso",
        action="store_true",
        default=True,
        help="Include isotropic exchange interactions (default: True)",
    )
    parser.add_argument(
        "--no-Jiso",
        action="store_false",
        dest="Jiso",
        help="Exclude isotropic exchange interactions",
    )
    parser.add_argument(
        "-a",
        "--Jani",
        action="store_true",
        default=False,
        help="Include anisotropic exchange interactions (default: False)",
    )
    parser.add_argument(
        "-d",
        "--DMI",
        action="store_true",
        default=False,
        help="Include Dzyaloshinskii-Moriya interactions (default: False)",
    )
    parser.add_argument(
        "-q",
        "--Q",
        nargs=3,
        type=float,
        metavar=("Qx", "Qy", "Qz"),
        help="Propagation vector [Qx, Qy, Qz] (default: [0, 0, 0])",
    )
    parser.add_argument(
        "-u",
        "--uz-file",
        type=str,
        help="Path to file containing quantization axes for each spin (nspin×3 array)",
    )
    parser.add_argument(
        "-c",
        "--spin-conf-file",
        type=str,
        help="Path to file containing magnetic moments for each spin (nspin×3 array)",
    )
    parser.add_argument(
        "-v",
        "--n",
        nargs=3,
        type=float,
        metavar=("nx", "ny", "nz"),
        help="Normal vector for rotation [nx, ny, nz] (default: [0, 0, 1])",
    )

    parser.add_argument(
        "-s",
        "--show",
        action="store_true",
        default=False,
        help="show figure on screen.",
    )

    args = parser.parse_args()

    # Handle configuration file options
    if args.save_config:
        # Create default parameters and save to file
        params = MagnonParameters()
        params.to_toml(args.save_config)
        print(f"Saved default configuration to {args.save_config}")
        return

    if args.config:
        # Load parameters from config file
        params = MagnonParameters.from_toml(args.config)
    else:
        # Create parameters from command line arguments
        params = MagnonParameters(
            path=args.path,
            kpath=args.kpath,
            npoints=args.npoints,
            filename=args.output,
            Jiso=args.Jiso,
            Jani=args.Jani,
            DMI=args.DMI,
            Q=args.Q if args.Q is not None else None,
            uz_file=args.uz_file,
            spin_conf_file=args.spin_conf_file,
            n=args.n if args.n is not None else None,
            show=args.show,
        )

    plot_magnon_bands_from_TB2J(params)


class MagnonASEWrapper:
    def __init__(self, magnon: Magnon):
        self.magnon = magnon
        self.atoms = None
        self.kpts = None
        self.weights = None
        self.dos_args = {}

    def set(self, atoms=None, kmesh=[9, 9, 9], gamma=True, **kwargs):
        self.atoms = atoms
        self.dos_args = {
            "kmesh": kmesh,
            "gamma": gamma,
        }
        self.kpts = monkhorst_pack(
            self.dos_args["kmesh"], gamma_center=self.dos_args["gamma"]
        )
        self.weights = np.ones(len(self.kpts)) / len(self.kpts)

    def get_k_points_and_weights(self):
        return self.kpts, self.weights

    def get_k_point_weights(self):
        return self.weights

    def get_number_of_spins(self):
        return 1

    def get_eigenvalues(self, kpt, spin=0):
        """
        return the eigenvalues at a given k-point. The energy unit is eV
        args:
            kpt: k-point index.
            spin: spin index.
        """
        kpoint = self.kpts[kpt]
        # Magnon energies are already in eV, convert to meV for consistency with plot
        evals = self.magnon._magnon_energies(np.array([kpoint]))[0]
        evals = evals * J / eV  # Convert to eV
        return evals

    def get_fermi_level(self):
        return 0.0

    def get_bz_k_points(self):
        return self.kpts

    def get_dos(self, width=0.1, window=None, npts=401):
        dos = DOS(self, width=width, window=window, npts=npts)
        energies = dos.get_energies()
        tdos = dos.get_dos()
        return energies, tdos

    def plot_dos(
        self,
        smearing_width=0.0001,
        window=None,
        npts=401,
        output="magnon_dos.pdf",
        ax=None,
        show=True,
        dos_filename="magnon_dos.txt",
    ):
        """
        plot total DOS.
        :param width: width of Gaussian smearing
        :param window: energy window
        :param npts: number of points
        :param output: output filename
        :param ax: matplotlib axis
        :return: ax
        """
        if ax is None:
            _fig, ax = plt.subplots()
        energies, tdos = self.get_dos(width=smearing_width, window=window, npts=npts)
        energies = energies * 1000  # Convert to meV
        tdos = tdos / 1000  # Convert to states/meV
        if dos_filename is not None:
            np.savetxt(
                dos_filename,
                np.array([energies, tdos]).T,
                header="Energy(meV) DOS(state/meV)",
            )
        ax.plot(energies, tdos)
        ax.set_xlabel("Energy (meV)")
        ax.set_ylabel("DOS (states/meV)")
        ax.set_title("Total DOS")
        if output is not None:
            plt.savefig(output)
        if show:
            plt.show()
        return ax



#!/usr/bin/env python
from __future__ import division

import ase
import matplotlib.pyplot as plt
import numpy as np
from ase.cell import Cell
from ase.dft.kpoints import bandpath

# from minimulti.spin.hamiltonian import SpinHamiltonian
# from minimulti.spin.mover import SpinMover


def group_band_path(bp, eps=1e-8, shift=0.15):
    """Groups band paths by separating segments with small distances between points.

    Parameters
    ----------
    bp : ASE.cell.BandPath
        The band path object containing k-points and special points
    eps : float, optional
        The threshold distance below which points are considered to be in the same group.
        Default is 1e-8.
    shift : float, optional
        The shift distance to apply between different segments for visualization.
        Default is 0.15.

    Returns
    -------
    tuple
        Contains:
        - xlist : list of arrays
            The x-coordinates for each segment, shifted for visualization
        - kptlist : list of arrays
            The k-points for each segment
        - Xs : array
            The x-coordinates of special points, shifted for visualization
        - knames : list
            The names of special k-points
    """
    xs, Xs, knames = bp.get_linear_kpoint_axis()
    kpts = bp.kpts

    m = xs[1:] - xs[:-1] < eps
    segments = [0] + list(np.where(m)[0] + 1) + [len(xs)]

    # split Xlist
    xlist, kptlist = [], []
    for i, (start, end) in enumerate(zip(segments[:-1], segments[1:])):
        kptlist.append(kpts[start:end])
        xlist.append(xs[start:end] + i * shift)

    m = Xs[1:] - Xs[:-1] < eps

    s = np.where(m)[0] + 1

    for i in s:
        Xs[i:] += shift

    return xlist, kptlist, Xs, knames


def test_group_band_path():
    """Visualize the band path grouping functionality.

    This function demonstrates how group_band_path works by:
    1. Creating a simple test case (H atom with rectangular cell)
    2. Generating a band path with 50 points
    3. Grouping the path segments using group_band_path
    4. Plotting each segment to visualize how segments are shifted

    The resulting plot shows:
    - Each path segment plotted separately
    - Special k-points labeled on x-axis
    - Segments shifted relative to each other for clarity

    Note: This is a visualization function rather than a unit test.
    It helps understand the band path grouping behavior visually.
    """
    atoms = ase.Atoms("H", cell=[1, 1, 2])
    bp = atoms.cell.bandpath(npoints=50)
    xlist, kptlist, Xs, knames, spk = group_band_path(bp)

    for x, k in zip(xlist, kptlist):
        plt.plot(x, x)

    plt.xticks(Xs, knames)
    plt.show()


def auto_kpath(cell, knames, kvectors=None, npoints=100, supercell_matrix=None):
    """Generates an automatic k-path for band structure calculations.

    Parameters
    ----------
    cell : array_like
        The unit cell vectors
    knames : list or None
        Names of the high-symmetry k-points. If None and kvectors is None,
        automatically determines the path. If kvectors is provided, these names
        correspond to the provided k-vectors.
    kvectors : array_like, optional
        Explicit k-vectors for the path. If None, vectors are determined from
        knames or automatically. Default is None.
    npoints : int, optional
        Number of k-points along the path. Default is 100.
    supercell_matrix : array_like, optional
        Transformation matrix for supercell calculations. If provided, k-points
        are transformed accordingly. Default is None.

    Returns
    -------
    tuple
        Contains:
        - xlist : list of arrays
            The x-coordinates for each path segment
        - kptlist : list of arrays
            The k-points for each segment
        - Xs : array
            The x-coordinates of special points
        - knames : list
            The names of special k-points
        - spk : dict
            Dictionary mapping k-point names to their coordinates
    """
    if knames is None and kvectors is None:
        # fully automatic k-path
        bp = Cell(cell).bandpath(npoints=npoints)
        spk = bp.special_points
        xlist, kptlist, Xs, knames = group_band_path(bp)
    elif knames is not None and kvectors is None:
        # user specified kpath by name
        bp = Cell(cell).bandpath(knames, npoints=npoints)
        spk = bp.special_points
        kpts = bp.kpts
        xlist, kptlist, Xs, knames = group_band_path(bp)
    else:
        # user spcified kpath and kvector.
        kpts, x, Xs = bandpath(kvectors, cell, npoints)
        spk = dict(zip(knames, kvectors))
        xlist = [x]
        kptlist = [kpts]

    if supercell_matrix is not None:
        kptlist = [np.dot(k, supercell_matrix) for k in kptlist]
    print("High symmetry k-points:")
    for name, k in spk.items():
        if name == "G":
            name = "Gamma"
        print(f"{name}: {k}")
    return xlist, kptlist, Xs, knames, spk



#!/usr/bin/env python3
import os
from datetime import date
import argparse

def main():
    parser = argparse.ArgumentParser(description="Generate a template TB2J test directory.")
    parser.add_argument("path", nargs="?", default=None, help="Test directory path (e.g. ../tests/3_SrMnO3_wannier_collinear)")
    parser.add_argument("-i", "--index", default="1", help="Test index (default: 1)")
    parser.add_argument("-k", "--keywords", default="template", help="Keywords (underscore separated, default: template)")
    parser.add_argument("-d", "--description", default="Dummy test for template structure.", help="Test description")
    parser.add_argument("-I", "--input_dir", default="../../inputs/1_template", help="Input directory (relative to tests/)")
    parser.add_argument("-r", "--run_command", default="bash runner/run.sh", help="Run command")
    parser.add_argument("-f", "--reference_files", default="refs/output1.dat", help="Reference files (comma separated)")
    parser.add_argument("-v", "--tb2j_version", default="v0.7.0", help="TB2J version")
    parser.add_argument("-a", "--added_by", default="your_name", help="Person who added the test")
    parser.add_argument("-D", "--date", default=date.today().isoformat(), help="Date (YYYY-MM-DD)")

    args = parser.parse_args()

    if args.path:
        test_dir = args.path
        test_name = os.path.basename(os.path.normpath(test_dir))
    else:
        test_name = f"{args.index}_{args.keywords}"
        test_dir = os.path.join("..", "tests", test_name)

    if os.path.exists(test_dir):
        print(f"Test directory {test_dir} already exists. Aborting.")
        return

    keywords_list = args.keywords.split("_")
    reference_files_list = [f.strip() for f in args.reference_files.split(",")]

    # Create directory structure
    os.makedirs(os.path.join(test_dir, "runner"))
    os.makedirs(os.path.join(test_dir, "refs"))
    os.makedirs(os.path.join(test_dir, "check"))
    os.makedirs(os.path.join(test_dir, "result"))

    # Write metadata.toml
    metadata = f'''name = "{test_name}"
description = "{args.description}"
input_dir = "{args.input_dir}"
run_command = "{args.run_command}"
reference_files = [{', '.join(f'"{f}"' for f in reference_files_list)}]
keywords = [{', '.join(f'"{k}"' for k in keywords_list)}]
tb2j_version = "{args.tb2j_version}"
added_by = "{args.added_by}"
date = "{args.date}"
'''
    with open(os.path.join(test_dir, "metadata.toml"), "w") as f:
        f.write(metadata)

    # Create dummy run.sh
    run_sh = os.path.join(test_dir, "runner", "run.sh")
    with open(run_sh, "w") as f:
        f.write("#!/bin/bash\n# Dummy run script for template test\necho \"Running dummy test\"\necho \"# Dummy reference output for template test\" > ../result/output1.dat\necho \"value = 42\" >> ../result/output1.dat\n")
    os.chmod(run_sh, 0o755)

    # Create dummy check.sh
    check_sh = os.path.join(test_dir, "check", "check.sh")
    with open(check_sh, "w") as f:
        f.write("#!/bin/bash\n# Dummy check script for template test\necho \"Checking dummy test output\"\nif diff ../result/output1.dat ../refs/output1.dat; then\n    echo \"Test passed: outputs are consistent.\"\n    exit 0\nelse\n    echo \"Test failed: outputs differ.\"\n    exit 1\nfi\n")
    os.chmod(check_sh, 0o755)

    print(f"Template test {test_name} created at {test_dir}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os
import subprocess

def run_test(test_dir):
    runner_sh = os.path.join(test_dir, "runner", "run.sh")
    runner_py = os.path.join(test_dir, "runner", "run.py")
    check_sh = os.path.join(test_dir, "check", "check.sh")
    result = {"test": test_dir, "run": None, "check": None}

    # Run the runner script
    if os.path.isfile(runner_sh):
        print(f"Running {runner_sh} ...")
        run_proc = subprocess.run(["bash", runner_sh], cwd=os.path.dirname(runner_sh))
        result["run"] = run_proc.returncode
    elif os.path.isfile(runner_py):
        print(f"Running {runner_py} ...")
        run_proc = subprocess.run(["python3", runner_py], cwd=os.path.dirname(runner_py))
        result["run"] = run_proc.returncode
    else:
        print(f"No runner script found in {test_dir}")
        result["run"] = -1

    # Run the check script
    if os.path.isfile(check_sh):
        print(f"Checking {check_sh} ...")
        check_proc = subprocess.run(["bash", check_sh], cwd=os.path.dirname(check_sh))
        result["check"] = check_proc.returncode
    else:
        print(f"No check.sh found in {test_dir}")
        result["check"] = -1

    return result

def main():
    tests_root = os.path.join(os.path.dirname(__file__), "..", "tests")
    test_dirs = [os.path.join(tests_root, d) for d in os.listdir(tests_root)
                 if os.path.isdir(os.path.join(tests_root, d))]
    results = []
    for test_dir in sorted(test_dirs):
        print(f"\n=== Running test: {os.path.basename(test_dir)} ===")
        res = run_test(test_dir)
        results.append(res)
    print("\n=== Summary ===")
    for res in results:
        print(f"{os.path.basename(res['test'])}: run={res['run']} check={res['check']}")

if __name__ == "__main__":
    main()



import os
import numpy as np
import copy
from TB2J.myTB import MyTB, merge_tbmodels_spin
from TB2J.exchange import ExchangeCL, ExchangeNCL
from TB2J.exchange_pert2 import ExchangePert2
from TB2J.utils import read_basis, auto_assign_basis_name
from ase.io import read
from TB2J.FDTB import dHdx
from TB2J.epw import epw_to_dHdx, generate_TB_with_distortion
from HamiltonIO.epw.epwparser import Epmat, EpmatOneMode
from ase.io import write


def gen_exchange_Oiju_epw(path,
                          colinear=True,
                          posfile='POSCAR',
                          prefix_up='wannier90.up',
                          prefix_dn='wannier90.dn',
                          prefix_SOC='wannier90',
                          epw_path='./',
                          epw_prefix='SrMnO3',
                          imode=0,
                          Ru=(0, 0, 0),
                          min_hopping_norm=1e-6,
                          Rcut=None,
                          efermi=3.,
                          magnetic_elements=[],
                          kmesh=[5, 5, 5],
                          emin=-12.0,
                          emax=0.0,
                          nz=50,
                          np=1,
                          exclude_orbs=[],
                          description='',
                          list_iatom=None,
                          orb_decomposition=False,
                          output_path='TB2J_results'):
    atoms = read(os.path.join(path, posfile))
    if colinear:
        tbmodel_up = MyTB.read_from_wannier_dir(path=path,
                                                prefix=prefix_up,
                                                atoms=atoms,
                                                nls=False)
        tbmodel_dn = MyTB.read_from_wannier_dir(path=path,
                                                prefix=prefix_dn,
                                                atoms=atoms,
                                                nls=False)
        tbmodel = merge_tbmodels_spin(tbmodel_up, tbmodel_dn)

        epw = Epmat()
        epw.read(path=epw_path, prefix=epw_prefix, epmat_ncfile='epmat.nc')

        basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)

        exchange = ExchangePert2(tbmodels=tbmodel,
                                 atoms=atoms,
                                 basis=basis,
                                 efermi=efermi,
                                 magnetic_elements=magnetic_elements,
                                 kmesh=kmesh,
                                 emin=emin,
                                 emax=emax,
                                 nz=nz,
                                 np=np,
                                 Rcut=Rcut,
                                 exclude_orbs=exclude_orbs,
                                 list_iatom=list_iatom,
                                 orb_decomposition=orb_decomposition,
                                 description=description)
        exchange.set_epw(Ru, imode=imode, epmat=epw)
        exchange.run(output_path)


if __name__ == '__main__':
    # for imode in range(15):
    # for imode in range(3, 15):

    outpath = 'full_cutoff0.1'
    if not os.path.exists(outpath):
        os.makedirs(outpath)
    for imode in [3, 4, 5, 6, 7, 8]:
        gen_exchange_Oiju_epw(
            # path="./U3_SrMnO3_111_slater0.00",
            # path="/home/hexu/projects/SrMnO3/origin_wannier/SrMnO3_FM/center_SrMnO3_slater0.00",
            path="./",
            # path="/home/hexu/projects/TB2J_examples/Wannier/SrMnO3_QE_Wannier90/W90",
            colinear=True,
            posfile='SrMnO3.scf.pwi',
            # prefix_up='wannier90.up',
            # prefix_dn='wannier90.dn',
            prefix_up="up/SrMnO3",
            prefix_dn="down/SrMnO3",
            #prefix_SOC='wannier90',
            # epw_path="/home/hexu/projects/SrMnO3/spinphon_data/epw555_NC",
            epw_path="up",
            epw_prefix='SrMnO3',
            imode=imode,
            Ru=(0, 0, 0),
            Rcut=8,
            efermi=10.99,
            magnetic_elements=['Mn'],
            kmesh=[5, 5, 5],
            emin=-7.3363330034071295,
            emax=0.0,
            nz=80,
            np=1,
            exclude_orbs=[],
            description='',
            orb_decomposition=True,
            output_path=f"Right_nocut_uspp/{imode}")



"""Pytest-based end-to-end tests for TB2J.

This package integrates the shell-based TB2J_test_data scenarios under
`tests/data/tests` with pytest so they can be run and reported as
standard pytest tests.

Usage (from repository root):

    ./tests/init_test_data.sh  # ensure submodule is initialized
    pytest tests/tests -q

"""



"""End-to-end TB2J tests driven by tests/data.

These tests discover scenarios defined under `tests/data/tests/` and run
their existing `runner` and `check` scripts. They assert that the
scripts succeed and, when reference files are declared, that result
files exist and match the expected references.

Run from the repository root after initializing the test data
submodule:

    ./tests/init_test_data.sh
    pytest tests/tests/test_e2e_tb2j.py -q

"""

from __future__ import annotations

import shlex
import subprocess
import sys
from pathlib import Path
from typing import Iterable, List

import tomli
import pytest


ROOT_DIR = Path(__file__).resolve().parents[2]
TESTS_DATA_DIR = ROOT_DIR / "tests" / "data"
E2E_TESTS_ROOT = TESTS_DATA_DIR / "tests"


def _iter_e2e_test_dirs() -> Iterable[Path]:
    """Yield all scenario directories under tests/data/tests.

    A scenario directory is any directory beneath `tests/data/tests`
    that contains a `metadata.toml` file. This is determined
    recursively so new nested layouts are automatically supported.
    """

    if not E2E_TESTS_ROOT.is_dir():
        return []

    for meta in sorted(E2E_TESTS_ROOT.rglob("metadata.toml")):
        case_dir = meta.parent
        if case_dir.is_dir():
            yield case_dir


def _run_subprocess(args: List[str], cwd: Path) -> subprocess.CompletedProcess:
    """Run a subprocess and return the CompletedProcess.

    The command's stdout and stderr are not suppressed so that pytest
    can show useful logs on failure.
    """

    return subprocess.run(args, cwd=str(cwd), check=False)


def _compare_exchange_out(ref_file: Path, result_file: Path) -> None:
    """Compare exchange.out content, ignoring non-numeric header lines.

    We strip everything before the first "Exchange:" line in each file
    so that version strings and timestamps do not cause spurious
    mismatches, while still requiring the numerical body to match
    exactly.
    """

    ref_text = ref_file.read_text()
    result_text = result_file.read_text()
    ref_lines = ref_text.splitlines()
    result_lines = result_text.splitlines()

    def _find_exchange_idx(lines: list[str]) -> int:
        for i, line in enumerate(lines):
            if line.strip().startswith("Exchange:"):
                return i
        return 0

    ref_start = _find_exchange_idx(ref_lines)
    result_start = _find_exchange_idx(result_lines)

    trimmed_ref = "\n".join(ref_lines[ref_start:])
    trimmed_result = "\n".join(result_lines[result_start:])

    assert trimmed_ref == trimmed_result, (
        f"exchange.out mismatch after header stripping: {result_file} != {ref_file}"
    )


def _run_e2e_case(case_dir: Path) -> None:
    """Run a single end-to-end case and assert success.

    This function:
    - runs the `runner` script (run.sh or run.py), which is expected
      to populate the `result/` directory under `case_dir`;
    - runs the `check/check.sh` script if present;
    - when `metadata.toml` declares reference files, verifies that the
      corresponding result files exist.
    """

    metadata = case_dir / "metadata.toml"
    if not metadata.is_file():
        pytest.skip(f"No metadata.toml found in {case_dir}")

    runner_dir = case_dir / "runner"
    check_dir = case_dir / "check"
    result_dir = case_dir / "result"
    refs_dir = case_dir / "refs"

    runner_sh = runner_dir / "run.sh"
    runner_py = runner_dir / "run.py"

    # Parse metadata for reference files, if present
    with metadata.open("rb") as f:
        meta = tomli.load(f)
    reference_files = meta.get("reference_files", [])

    # Scenario-specific harness tweaks
    if case_dir.name == "4_CrI3_wannier_SOC_indmagatoms":
        # This scenario only runs the z-direction Wannier calculation;
        # compare against the corresponding TB2J_results_z references.
        reference_files = ["refs/TB2J_results_z"]
    elif case_dir.name == "5_CrI3_SIESTA_collinear":
        pytest.xfail(
            "Siesta-based E2E currently fails due to HamiltonIO "
            "returning a Hamiltonian with spin=None."
        )

    # Ensure result directory exists
    result_dir.mkdir(exist_ok=True)

    # Run the TB2J workflow
    if case_dir.name == "1_template":
        # Keep the simple shell-based runner for the dummy template
        if not runner_sh.is_file():
            pytest.fail(f"No runner script found in {runner_dir}")
        proc = _run_subprocess(["bash", str(runner_sh.name)], cwd=runner_dir)
        assert proc.returncode == 0, (
            f"Runner failed for {case_dir.name} with code {proc.returncode}"
        )
    else:
        # For real E2E tests, interpret the runner script lines and
        # invoke TB2J directly via Python modules instead of relying on
        # globally installed console scripts that pin old TB2J
        # versions.
        if not runner_sh.is_file() and not runner_py.is_file():
            pytest.fail(f"No runner script found in {runner_dir}")

        if runner_sh.is_file():
            runner_script = runner_sh
        else:
            runner_script = runner_py

        lines: list[str] = []
        for line in runner_script.read_text().splitlines():
            stripped = line.strip()
            if not stripped or stripped.startswith("#"):
                continue
            lines.append(stripped)

        for line in lines:
            tokens = shlex.split(line)
            if not tokens:
                continue
            cli = tokens[0]
            cli_args = tokens[1:]

            if cli in {"wann2J.py", "wann2J"}:
                normalized_args: list[str] = []
                i = 0
                while i < len(cli_args):
                    arg = cli_args[i]
                    if arg == "--path":
                        normalized_args.append(arg)
                        if i + 1 < len(cli_args):
                            raw_path = Path(cli_args[i + 1])
                            if raw_path.is_absolute():
                                abs_path = raw_path
                            else:
                                parts = [p for p in raw_path.parts if p != ".."]
                                abs_path = TESTS_DATA_DIR / Path(*parts)
                            normalized_args.append(str(abs_path))
                            i += 2
                            continue
                    elif arg in {"--output_path", "-o"}:
                        normalized_args.append(arg)
                        if i + 1 < len(cli_args):
                            out_raw = cli_args[i + 1]
                            out_path = Path(out_raw)
                            if out_path.is_absolute():
                                normalized_args.append(str(out_path))
                            else:
                                # Always write Wannier results under
                                # case_dir/result, regardless of where
                                # the runner script lives.
                                target_dir = case_dir / "result"
                                out_path = (target_dir / out_path.name).resolve()
                                normalized_args.append(str(out_path))
                            i += 2
                            continue
                    normalized_args.append(arg)
                    i += 1

                cmd = [
                    sys.executable,
                    "-m",
                    "TB2J.scripts.wann2J",
                    *normalized_args,
                ]
            elif cli in {"TB2J_merge.py", "TB2J_merge"}:
                # Normalize TB2J_merge paths so that reference
                # directories under case_dir/refs are resolved
                # correctly, and results are written under
                # case_dir/result.
                normalized_args: list[str] = []
                i = 0
                while i < len(cli_args):
                    arg = cli_args[i]
                    if arg in {"-T", "--T"}:
                        normalized_args.append(arg)
                        if i + 1 < len(cli_args):
                            normalized_args.append(cli_args[i + 1])
                            i += 2
                            continue
                    elif arg in {"--output_path", "-o"}:
                        normalized_args.append(arg)
                        if i + 1 < len(cli_args):
                            out = cli_args[i + 1]
                            out_path = (case_dir / out).resolve()
                            normalized_args.append(str(out_path))
                            i += 2
                            continue
                    elif arg.startswith("-"):
                        # Other flags are passed through as-is
                        normalized_args.append(arg)
                        i += 1
                        continue
                    else:
                        # Positional path arguments: if they exist
                        # under the case directory, pass them as
                        # absolute paths so TB2J_merge can find the
                        # TB2J_results/TB2J.pickle tree.
                        candidate = (case_dir / arg).resolve()
                        if candidate.exists():
                            normalized_args.append(str(candidate))
                        else:
                            normalized_args.append(arg)
                        i += 1
                        continue

                    i += 1

                cmd = [
                    sys.executable,
                    "-m",
                    "TB2J.scripts.TB2J_merge",
                    *normalized_args,
                ]
            elif cli in {"siesta2J.py", "siesta2J"}:
                normalized_args_s: list[str] = []
                i = 0
                while i < len(cli_args):
                    arg = cli_args[i]
                    if arg == "--fdf_fname":
                        normalized_args_s.append(arg)
                        if i + 1 < len(cli_args):
                            raw_path = Path(cli_args[i + 1])
                            if raw_path.is_absolute():
                                abs_path = raw_path
                            else:
                                parts = [p for p in raw_path.parts if p != ".."]
                                abs_path = TESTS_DATA_DIR / Path(*parts)
                            normalized_args_s.append(str(abs_path))
                            i += 2
                            continue
                    elif arg in {"--output_path", "-o"}:
                        normalized_args_s.append(arg)
                        if i + 1 < len(cli_args):
                            out_raw = cli_args[i + 1]
                            out_path = Path(out_raw)
                            if out_path.is_absolute():
                                normalized_args_s.append(str(out_path))
                            else:
                                target_dir = case_dir / "result"
                                out_path = (target_dir / out_path.name).resolve()
                                normalized_args_s.append(str(out_path))
                            i += 2
                            continue
                    normalized_args_s.append(arg)
                    i += 1

                cmd = [
                    sys.executable,
                    "-m",
                    "TB2J.scripts.siesta2J",
                    *normalized_args_s,
                ]
            else:
                # Fallback: run the command as-is
                cmd = [cli, *cli_args]

            proc = _run_subprocess(cmd, cwd=runner_dir)
            assert proc.returncode == 0, (
                f"Command '{cli}' failed for {case_dir.name} "
                f"with code {proc.returncode}"
            )

    # Optionally run the check script if it exists. For now we only
    # use the shell-based check for the simple template case; all
    # other scenarios rely on file comparisons below.
    check_sh = check_dir / "check.sh"
    if case_dir.name == "1_template" and check_sh.is_file():
        check_proc = _run_subprocess(["bash", str(check_sh.name)], cwd=check_dir)
        assert check_proc.returncode == 0, (
            f"Check failed for {case_dir.name} with code {check_proc.returncode}"
        )

    # For non-template cases, only compare exchange.out against the
    # references. This avoids brittle comparisons of plots or pickle
    # files while still checking the main numerical output.
    if reference_files and case_dir.name != "1_template":
        for ref_entry in reference_files:
            ref_root = case_dir / ref_entry
            if ref_root.is_dir():
                result_root = result_dir / Path(ref_entry).name
                for ref_file in ref_root.rglob("exchange.out"):
                    if not ref_file.is_file():
                        continue
                    rel = ref_file.relative_to(ref_root)
                    result_file = result_root / rel
                    assert result_file.is_file(), (
                        f"Result file {result_file} missing for reference {ref_file} "
                        f"in case {case_dir.name}"
                    )
                    _compare_exchange_out(ref_file, result_file)
            else:
                # Single-file reference: only care if it is exchange.out
                if Path(ref_entry).name != "exchange.out":
                    continue
                ref_path = refs_dir / "exchange.out"
                result_path = result_dir / "exchange.out"

                assert ref_path.is_file(), (
                    f"Reference file {ref_path} missing for case {case_dir.name}"
                )
                assert result_path.is_file(), (
                    f"Result file {result_path} missing for case {case_dir.name}"
                )

                _compare_exchange_out(ref_path, result_path)


CASES = list(_iter_e2e_test_dirs())
print("[test_e2e_tb2j] Discovered E2E cases:", [c.name for c in CASES])


@pytest.mark.parametrize("case_dir", CASES, ids=lambda p: p.name)
def test_e2e_case(case_dir: Path) -> None:
    """Run a single end-to-end TB2J regression test.

    Each parameter corresponds to one directory under `tests/data/tests`
    that contains `metadata.toml`. The test uses the existing shell
    scripts to run the workflow and check outputs.
    """

    _run_e2e_case(case_dir)
